@@ -5,10 +5,12 @@
 #include <stdio.h>
 #include <ctype.h>
 
-#include "ntp_fp.h"
+//#include "ntpd.h"	/* Only for DPRINTF */
+//#include "ntp_fp.h"
 #include "ntp.h"
 #include "ntp_syslog.h"
 #include "ntp_stdlib.h"
+#include "ntp_keyacc.h"
 
 #ifdef OPENSSL
 #include "openssl/objects.h"
@@ -36,7 +38,7 @@ nexttok(
 	 */
 	while (*cp == ' ' || *cp == '\t')
 		cp++;
-	
+
 	/*
 	 * Save this and space to end of token
 	 */
@@ -44,19 +46,19 @@ nexttok(
 	while (*cp != '\0' && *cp != '\n' && *cp != ' '
 	       && *cp != '\t' && *cp != '#')
 		cp++;
-	
+
 	/*
 	 * If token length is zero return an error, else set end of
 	 * token to zero and return start.
 	 */
 	if (starttok == cp)
 		return NULL;
-	
+
 	if (*cp == ' ' || *cp == '\t')
 		*cp++ = '\0';
 	else
 		*cp = '\0';
-	
+
 	*str = cp;
 	return starttok;
 }
@@ -85,6 +87,7 @@ static void log_maybe(u_int*, const char*, ...) NTP_PRINTF(2, 3);
 typedef struct keydata KeyDataT;
 struct keydata {
 	KeyDataT *next;		/* queue/stack link		*/
+	KeyAccT  *keyacclist;	/* key access list		*/
 	keyid_t   keyid;	/* stored key ID		*/
 	u_short   keytype;	/* stored key type		*/
 	u_short   seclen;	/* length of secret		*/
@@ -98,13 +101,33 @@ log_maybe(
 	...)
 {
 	va_list ap;
-	if (++(*pnerr) <= nerr_loglimit) {
+	if ((NULL == pnerr) || (++(*pnerr) <= nerr_loglimit)) {
 		va_start(ap, fmt);
 		mvsyslog(LOG_ERR, fmt, ap);
 		va_end(ap);
 	}
 }
 
+static void
+free_keydata(
+	KeyDataT *node
+	)
+{
+	KeyAccT *kap;
+
+	if (node) {
+		while (node->keyacclist) {
+			kap = node->keyacclist;
+			node->keyacclist = kap->next;
+			free(kap);
+		}
+
+		/* purge secrets from memory before free()ing it */
+		memset(node, 0, sizeof(*node) + node->seclen);
+		free(node);
+	}
+}
+
 /*
  * authreadkeys - (re)read keys from a file.
  */
@@ -119,12 +142,12 @@ authreadkeys(
 	keyid_t	keyno;
 	int	keytype;
 	char	buf[512];		/* lots of room for line */
-	u_char	keystr[32];		/* Bug 2537 */
+	u_char	keystr[AUTHPWD_MAXSECLEN];
 	size_t	len;
-	size_t	j;
 	u_int   nerr;
 	KeyDataT *list = NULL;
 	KeyDataT *next = NULL;
+
 	/*
 	 * Open file.  Complain and return if it can't be opened.
 	 */
@@ -148,12 +171,12 @@ authreadkeys(
 		token = nexttok(&line);
 		if (token == NULL)
 			continue;
-		
+
 		/*
 		 * First is key number.  See if it is okay.
 		 */
 		keyno = atoi(token);
-		if (keyno == 0) {
+		if (keyno < 1) {
 			log_maybe(&nerr,
 				  "authreadkeys: cannot change key %s",
 				  token);
@@ -177,39 +200,48 @@ authreadkeys(
 				  keyno);
 			continue;
 		}
+
+		/* We want to silently ignore keys where we do not
+		 * support the requested digest type. OTOH, we want to
+		 * make sure the file is well-formed.  That means we
+		 * have to process the line completely and have to
+		 * finally throw away the result... This is a bit more
+		 * work, but it also results in better error detection.
+		 */
 #ifdef OPENSSL
 		/*
-		 * The key type is the NID used by the message digest 
+		 * The key type is the NID used by the message digest
 		 * algorithm. There are a number of inconsistencies in
 		 * the OpenSSL database. We attempt to discover them
 		 * here and prevent use of inconsistent data later.
 		 */
 		keytype = keytype_from_text(token, NULL);
 		if (keytype == 0) {
-			log_maybe(&nerr,
+			log_maybe(NULL,
 				  "authreadkeys: invalid type for key %d",
 				  keyno);
-			continue;
-		}
-		if (EVP_get_digestbynid(keytype) == NULL) {
-			log_maybe(&nerr,
+#  ifdef ENABLE_CMAC
+		} else if (NID_cmac != keytype &&
+				EVP_get_digestbynid(keytype) == NULL) {
+			log_maybe(NULL,
 				  "authreadkeys: no algorithm for key %d",
 				  keyno);
-			continue;
+			keytype = 0;
+#  endif /* ENABLE_CMAC */
 		}
 #else	/* !OPENSSL follows */
-
 		/*
 		 * The key type is unused, but is required to be 'M' or
 		 * 'm' for compatibility.
 		 */
 		if (!(*token == 'M' || *token == 'm')) {
-			log_maybe(&nerr,
+			log_maybe(NULL,
 				  "authreadkeys: invalid type for key %d",
 				  keyno);
-			continue;
+			keytype = 0;
+		} else {
+			keytype = KEY_TYPE_MD5;
 		}
-		keytype = KEY_TYPE_MD5;
 #endif	/* !OPENSSL */
 
 		/*
@@ -225,59 +257,118 @@ authreadkeys(
 			continue;
 		}
 		next = NULL;
-		len = strlen(token);
-		if (len <= 20) {	/* Bug 2537 */
-			next = emalloc(sizeof(KeyDataT) + len);
-			next->keyid   = keyno;
-			next->keytype = keytype;
-			next->seclen  = len;
-			memcpy(next->secbuf, token, len);
-		} else {
-			static const char hex[] = "0123456789abcdef";
-			u_char	temp;
-			char	*ptr;
-			size_t	jlim;
-
-			jlim = min(len, 2 * sizeof(keystr));
-			for (j = 0; j < jlim; j++) {
-				ptr = strchr(hex, tolower((unsigned char)token[j]));
-				if (ptr == NULL)
-					break;	/* abort decoding */
-				temp = (u_char)(ptr - hex);
-				if (j & 1)
-					keystr[j / 2] |= temp;
-				else
-					keystr[j / 2] = temp << 4;
-			}
-			if (j < jlim) {
+		len = authdecodepw(keystr, sizeof(keystr), token, AUTHPWD_UNSPEC);
+		if (len > sizeof(keystr)) {
+			switch (errno) {
+			case ENOMEM:
 				log_maybe(&nerr,
-					  "authreadkeys: invalid hex digit for key %d",
+					  "authreadkeys: passwd too long for key %d",
 					  keyno);
-				continue;
+				break;
+			case EINVAL:
+				log_maybe(&nerr,
+					  "authreadkeys: passwd has bad char for key %d",
+					  keyno);
+				break;
+			default:
+				log_maybe(&nerr,
+					  "authreadkeys: unknown errno %d for key %d",
+					  errno, keyno);
+				break;
 			}
-			len = jlim/2; /* hmmmm.... what about odd length?!? */
-			next = emalloc(sizeof(KeyDataT) + len);
-			next->keyid   = keyno;
-			next->keytype = keytype;
-			next->seclen  = len;
-			memcpy(next->secbuf, keystr, len);
+			continue;
 		}
-		if (next) {
-			next->next = list;
-			list = next;
+		next = emalloc(sizeof(KeyDataT) + len);
+		next->keyacclist = NULL;
+		next->keyid   = keyno;
+		next->keytype = keytype;
+		next->seclen  = len;
+		memcpy(next->secbuf, keystr, len);
+
+		token = nexttok(&line);
+		if (token != NULL) {	/* A comma-separated IP access list */
+			char *tp = token;
+
+			while (tp) {
+				char *i;
+				char *snp;	/* subnet text pointer */
+				unsigned int snbits;
+				sockaddr_u addr;
+
+				i = strchr(tp, (int)',');
+				if (i) {
+					*i = '\0';
+				}
+				snp = strchr(tp, (int)'/');
+				if (snp) {
+					char *sp;
+
+					*snp++ = '\0';
+					snbits = 0;
+					sp = snp;
+
+					while (*sp != '\0') {
+						if (!isdigit((unsigned char)*sp))
+						    break;
+						if (snbits > 1000)
+						    break;	/* overflow */
+						snbits = 10 * snbits + (*sp++ - '0');       /* ascii dependent */
+					}
+					if (*sp != '\0') {
+						log_maybe(&nerr,
+							  "authreadkeys: Invalid character in subnet specification for <%s/%s> in key %d",
+							  sp, snp, keyno);
+						goto nextip;
+					}
+				} else {
+					snbits = UINT_MAX;
+				}
+
+				if (is_ip_address(tp, AF_UNSPEC, &addr)) {
+					/* Make sure that snbits is valid for addr */
+				    if ((snbits < UINT_MAX) &&
+					( (IS_IPV4(&addr) && snbits > 32) ||
+					  (IS_IPV6(&addr) && snbits > 128))) {
+						log_maybe(NULL,
+							  "authreadkeys: excessive subnet mask <%s/%s> for key %d",
+							  tp, snp, keyno);
+				    }
+				    next->keyacclist = keyacc_new_push(
+					next->keyacclist, &addr, snbits);
+				} else {
+					log_maybe(&nerr,
+						  "authreadkeys: invalid IP address <%s> for key %d",
+						  tp, keyno);
+				}
+
+			nextip:
+				if (i) {
+					tp = i + 1;
+				} else {
+					tp = 0;
+				}
+			}
 		}
+
+		/* check if this has to be weeded out... */
+		if (0 == keytype) {
+			free_keydata(next);
+			next = NULL;
+			continue;
+		}
+
+		INSIST(NULL != next);
+		next->next = list;
+		list = next;
 	}
 	fclose(fp);
-	if (nerr > nerr_maxlimit) {
-		msyslog(LOG_ERR,
-			"authreadkeys: rejecting file '%s' after %u errors (emergency break)",
-			file, nerr);
-		goto onerror;
-	}
 	if (nerr > 0) {
+		const char * why = "";
+		if (nerr > nerr_maxlimit)
+			why = " (emergency break)";
 		msyslog(LOG_ERR,
-			"authreadkeys: rejecting file '%s' after %u error(s)",
-			file, nerr);
+			"authreadkeys: rejecting file '%s' after %u error(s)%s",
+			file, nerr, why);
 		goto onerror;
 	}
 
@@ -287,10 +378,9 @@ authreadkeys(
 	while (NULL != (next = list)) {
 		list = next->next;
 		MD5auth_setkey(next->keyid, next->keytype,
-			       next->secbuf, next->seclen);
-		/* purge secrets from memory before free()ing it */
-		memset(next, 0, sizeof(*next) + next->seclen);
-		free(next);
+			       next->secbuf, next->seclen, next->keyacclist);
+		next->keyacclist = NULL; /* consumed by MD5auth_setkey */
+		free_keydata(next);
 	}
 	return (1);
 
@@ -298,9 +388,7 @@ authreadkeys(
 	/* Mop up temporary storage before bailing out. */
 	while (NULL != (next = list)) {
 		list = next->next;
-		/* purge secrets from memory before free()ing it */
-		memset(next, 0, sizeof(*next) + next->seclen);
-		free(next);
+		free_keydata(next);
 	}
 	return (0);
 }
