@@ -5,10 +5,20 @@
 
 #include "unity.h"
 
+void setUp(void);
 void test_RegularTime(void);
 void test_CurrentTime(void);
 
 
+void
+setUp(void)
+{
+	init_lib();
+
+	return;
+}
+
+
 void
 test_RegularTime(void)
 {
